/***************************************************************************\
* LRAHEADERSTART
*
*                   Copyright (C) 2004 Georgia SoftWorks.
*                   All rights reserved.
*
* LRAHEADEREND
\***************************************************************************/


#include <windows.h>

#include <winsock.h>
#include <stdio.h>
#include <stdlib.h>

#include "intrface.h"

#pragma comment(lib, "wsock32.lib")

/*
 * Intrface.c
 *
 *  To compile type cl intrface.c
 */


/* Handles for interface with GS Telnet Server */
static HANDLE s_hRlsIEvent = NULL;
static HANDLE s_hGetIEvent = NULL;
static HANDLE s_hRlsOEvent = NULL;
static HANDLE s_hGetOEvent = NULL;
static HANDLE s_hAckEvent  = NULL;
static HANDLE s_hPIRequest = NULL;

static BOOL s_appInitEvents(void);
static void s_appUninitEvents(void);
static LPVOID s_appOpenSharedMemory(DWORD dwSize, LPCSTR strName);
static SOCKET s_appGetSocket(GSW_INTERFACE1 *pPIData);


void main(void)
{
    WSADATA Data;
    int nStatus;
    SOCKET hSocket;
    char *szHSocket;
    char *szAgentPID;
    char ch;
    char szBuf[MAX_PATH];
    GSW_INTERFACE1 *pPIData = NULL;
    HANDLE hAgnt = NULL;
    BOOL bFirst = TRUE;

    __try
    {
	    //Old way of getting the socket, simple and still supported
	    //but fails after session is reconnected
        /* Get socket handle from environment */
        //szHSocket = getenv("GWTN_HSOCKET");
        //if (szHSocket == NULL)
        //{
        //    printf("Environment not set (GWTN_HSOCKET)");
        //    __leave;
        //}
        //hSocket = (SOCKET)atoi(szHSocket);

        BOOL bDone = FALSE;

        /*
         * Get name of the shared memory assigned for communication
         * with the Session Manager (GS_Agnt.exe
         */
        char *szSMPIData = getenv("GWTN_INTERFACE1");
        if (szSMPIData == NULL)
        {
            //fatal
            printf("Environment not set (GWTN_INTERFACE1).");
            __leave;
        }

        /*
         * Get pointer to the shared memory area. This call is made once
         * during application's lifetime
         */
    	if ((pPIData =  (GSW_INTERFACE1*)s_appOpenSharedMemory(S_GSW_INTERFACE1,
                                            szSMPIData)) == NULL)
	    {
            //fatal
		    printf("\nFailed to get shared memory for programmatic interface.");
    		__leave;
	    }

        
        if (pPIData->dwSize < S_GSW_INTERFACE1)
	    {
            //fatal
		    printf("\nIncompatible version of GSW UTS.");
    		__leave;
	    }

        /* Initialize interface events */
        if (!s_appInitEvents())
        {
            //fatal
            printf("\nCould not open events.");
            __leave;
        }

        /* Initialize WINSOCK */
        nStatus = WSAStartup(MAKEWORD(1, 1), &Data);
        if (nStatus != 0)
        {
            //fatal
            printf("\nAttempt to run socket app without tcp/ip support");
            __leave;
        }

        while (!bDone)
        {
            if (!bFirst)
            {
                DWORD dwCnt = 0;
                // We are here because a network error occured
                while (!pPIData->dwReconnected) 
                {
                    printf("\nWaiting for the Session Manager to reconnect.");
                    Sleep(5000);
                    dwCnt++;

                    /*
                     * If waiting here for too long we could also check
                     *
                     * pPIData->dwSuspended
                     *
                     * to make sure the session got suspended. Otherwise we
                     * may sit in this loop forever (unless the Session Manager
                     * kills our application during its cleanup procedures).
                     */
                    if (dwCnt > 5 && pPIData->dwSuspended == 0)
                        __leave;
                }
                // Clear the flag (this flag is set for us by Session Manager
                pPIData->dwReconnected = 0;
            }
            else
                bFirst = FALSE;


            /* Get socket handle */
            hSocket = s_appGetSocket(pPIData);
            if (hSocket == INVALID_SOCKET)
            {
                //fatal
                printf("\nInvalid socket returned.");
                __leave;
            }
            printf("\nHandle is %x.", hSocket);

            /*
             *   Take over input from socket and wait for acknowledgement
             *   from server
             */
            SetEvent(s_hGetIEvent);
            if (WaitForSingleObject(s_hAckEvent, INFINITE) != WAIT_OBJECT_0)
                __leave;

            /*
             *  Take over output to socket and wait for acknowledgement
             *  from server
             */
            SetEvent(s_hGetOEvent);
            if (WaitForSingleObject(s_hAckEvent, INFINITE) != WAIT_OBJECT_0)
                __leave;

            /*
             *  Wait for chars and echo them until 'q' pressed
             */
            if (send(hSocket, "\n\rEnter text (I will echo):\n\r", 29, 0) != 29)
                printf("\nSend failed, error %u.", WSAGetLastError());
            while (recv(hSocket, &ch, 1, 0) == 1)
            {
                if (send(hSocket, &ch, 1, 0) != 1)
                    break;
                if (ch == 'q')
                {
                    bDone = TRUE;
                    break;
                }
            }
            
            /* Network error occured, close the socket to avoid handle leak */
            closesocket(hSocket);

            /*
             *   Release control to GSW UTS to handle the error and possibly
             *   suspend the session (very important).
             */
            SetEvent(s_hRlsIEvent);
            if (WaitForSingleObject(s_hAckEvent, INFINITE) != WAIT_OBJECT_0)
                __leave;
            SetEvent(s_hRlsOEvent);
            if (WaitForSingleObject(s_hAckEvent, INFINITE) != WAIT_OBJECT_0)
                __leave;
        }

    }
    __finally
    {
        /*
         *   Release control to GSW UTS, server is protected against
         *   releasing control too many times.
         */
        if (s_hRlsIEvent != NULL)
            SetEvent(s_hRlsIEvent);
        if (s_hAckEvent != NULL)
            WaitForSingleObject(s_hAckEvent, INFINITE);
        if (s_hRlsOEvent != NULL)
            SetEvent(s_hRlsOEvent);
        if (s_hAckEvent != NULL)
            WaitForSingleObject(s_hAckEvent, INFINITE);

        /* Clean up event handles */
        s_appUninitEvents();

        /* Close Winsock                                          */
        WSACleanup();
    }
}


static BOOL s_appInitEvents(void)
{
    if ((s_hRlsIEvent = OpenEvent(EVENT_ALL_ACCESS,
                                  FALSE,
                                  getenv("GWTN_RLS_I"))) == NULL)
        return FALSE;
    if ((s_hRlsOEvent = OpenEvent(EVENT_ALL_ACCESS,
                                  FALSE,
                                  getenv("GWTN_RLS_O"))) == NULL)
        return FALSE;
    if ((s_hGetIEvent = OpenEvent(EVENT_ALL_ACCESS,
                                  FALSE,
                                  getenv("GWTN_GET_I"))) == NULL)
        return FALSE;
    if ((s_hGetOEvent = OpenEvent(EVENT_ALL_ACCESS,
                                  FALSE,
                                  getenv("GWTN_GET_O"))) == NULL)
        return FALSE;
    if ((s_hAckEvent = OpenEvent(EVENT_ALL_ACCESS,
                                 FALSE,
                                 getenv("GWTN_ACK"))) == NULL)
        return FALSE;
    if ((s_hPIRequest = OpenEvent(EVENT_ALL_ACCESS,
                                  FALSE,
                                  getenv("GWTN_PI_Request"))) == NULL)
        return FALSE;
                        

    return TRUE;
}


static void s_appUninitEvents(void)
{
    if (s_hRlsIEvent != NULL)
        CloseHandle(s_hRlsIEvent);
    if (s_hGetIEvent != NULL)
        CloseHandle(s_hGetIEvent);
    if (s_hRlsOEvent != NULL)
        CloseHandle(s_hRlsOEvent);
    if (s_hGetOEvent != NULL)
        CloseHandle(s_hGetOEvent);
    if (s_hAckEvent != NULL)
        CloseHandle(s_hAckEvent);
    if (s_hPIRequest != NULL)
        CloseHandle(s_hPIRequest); 
}

static LPVOID s_appOpenSharedMemory(DWORD dwSize, LPCSTR strName)
{
    HANDLE hMapFile;
    LPVOID lpMapAddress;

    hMapFile = OpenFileMapping(FILE_MAP_ALL_ACCESS, // Read/write permission. 
                               FALSE,               // Do not inherit the name
                               strName);            // of the mapping object. 
     
    if (hMapFile == NULL) 
    { 
        printf("\nOpenFileMapping() failed, error %u.", GetLastError());
        printf("\nReturning error from s_appOpenSharedMemory().");
        return NULL;
    } 
     
    lpMapAddress = MapViewOfFile(hMapFile, // Handle to mapping object. 
                                 FILE_MAP_ALL_ACCESS,               // Read/write permission. 
                                 0,                                 // Max. object size. 
                                 0,                                 // Size of hFile. 
                                 dwSize);                           // Map entire file. 
     
    if (lpMapAddress == NULL) 
    { 
        printf("\nMapViewOfFile() failed, error %u.", GetLastError());
        printf("\nReturning error from s_appOpenSharedMemory().");
    }
    else
        printf("\nReturning valid pointer from s_appOpenSharedMemory.");

    return lpMapAddress;
}

static SOCKET s_appGetSocket(GSW_INTERFACE1 *pPIData)
{
    pPIData->dwRequestorPID = GetCurrentProcessId();
    pPIData->dwRequestID = GSW_REQUEST_CLNT_SOCKET;

    /*
     * Wait for Session Manager to prepare the socket handle for our program
     */
    SetEvent(s_hPIRequest);
    if (WaitForSingleObject(s_hAckEvent, INFINITE) != WAIT_OBJECT_0)
    {
        printf("\nReturning error from s_appGetSocket().");
        return INVALID_SOCKET;
    }
    if (pPIData->dwRequestResult != ERROR_SUCCESS)
    {
        printf("\nSession Manager returned error %u.", pPIData->dwRequestResult);
        printf("\nReturning error from s_appGetSocket().");
        return INVALID_SOCKET;
    }

    printf("\nReturning valid socket in s_appGetSocket().");
    return pPIData->hClientSocket;
}
